@@ -7,7 +7,7 @@ import isort.literal
 from isort.settings import DEFAULT_CONFIG, Config
 
 from . import output, parse
-from .exceptions import FileSkipComment
+from .exceptions import ExistingSyntaxErrors, FileSkipComment
 from .format import format_natural, remove_whitespace
 from .settings import FILE_SKIP_COMMENTS
 
@@ -24,6 +24,7 @@ CODE_SORT_COMMENTS = (
     "# isort: unique-tuple",
     "# isort: assignments",
 )
+LITERAL_TYPE_MAPPING = {"(": "tuple", "[": "list", "{": "set"}
 
 
 def process(
@@ -72,6 +73,9 @@ def process(
     stripped_line: str = ""
     end_of_file: bool = False
     verbose_output: List[str] = []
+    lines_before: List[str] = []
+    is_reexport: bool = False
+    sort_section_pointer: int = 0
 
     if config.float_to_top:
         new_input = ""
@@ -129,6 +133,8 @@ def process(
                 line_separator = "\n"
 
             if code_sorting and code_sorting_section:
+                if is_reexport:
+                    output_stream.seek(sort_section_pointer, 0)
                 sorted_code = textwrap.indent(
                     isort.literal.assignment(
                         code_sorting_section,
@@ -144,7 +150,7 @@ def process(
                     line_separator=line_separator,
                     ignore_whitespace=config.ignore_whitespace,
                 )
-                output_stream.write(sorted_code)
+                sort_section_pointer += output_stream.write(sorted_code)
         else:
             stripped_line = line.strip()
             if stripped_line and not line_separator:
@@ -224,6 +230,14 @@ def process(
                     code_sorting = stripped_line.split("isort: ")[1].strip()
                     code_sorting_indent = line[: -len(line.lstrip())]
                     not_imports = True
+                elif config.sort_reexports and stripped_line.startswith("__all__"):
+                    _, rhs = stripped_line.split("=")
+                    code_sorting = LITERAL_TYPE_MAPPING.get(rhs.lstrip()[0], "tuple")
+                    code_sorting_indent = line[: -len(line.lstrip())]
+                    not_imports = True
+                    code_sorting_section += line
+                    is_reexport = True
+                    sort_section_pointer -= len(line)
                 elif code_sorting:
                     if not stripped_line:
                         sorted_code = textwrap.indent(
@@ -241,11 +255,14 @@ def process(
                             line_separator=line_separator,
                             ignore_whitespace=config.ignore_whitespace,
                         )
-                        output_stream.write(sorted_code)
+                        if is_reexport:
+                            output_stream.seek(sort_section_pointer, 0)
+                        sort_section_pointer += output_stream.write(sorted_code)
                         not_imports = True
                         code_sorting = False
                         code_sorting_section = ""
                         code_sorting_indent = ""
+                        is_reexport = False
                     else:
                         code_sorting_section += line
                         line = ""
@@ -285,6 +302,10 @@ def process(
                         else:
                             while ")" not in stripped_line:
                                 line = input_stream.readline()
+
+                                if not line:  # end of file without closing parenthesis
+                                    raise ExistingSyntaxErrors("Parenthesis is not closed")
+
                                 stripped_line = line.strip().split("#")[0]
                                 import_statement += line
 
@@ -304,7 +325,10 @@ def process(
                             or " cimport " in import_statement
                             or " cimport*" in import_statement
                             or " cimport(" in import_statement
-                            or ".cimport" in import_statement
+                            or (
+                                ".cimport" in import_statement
+                                and "cython.cimports" not in import_statement
+                            )  # Allow pure python imports. See #2062
                         ):
                             cimport_statement = True
 
@@ -332,7 +356,18 @@ def process(
                 else:
                     not_imports = True
 
+        sort_section_pointer += len(line)
+
         if not_imports:
+
+            if not was_in_quote and config.lines_before_imports > -1:
+                if line.strip() == "":
+                    lines_before += line
+                    continue
+                if not import_section:
+                    output_stream.write("".join(lines_before))
+                lines_before = []
+
             raw_import_section: str = import_section
             if (
                 add_imports
